@@ -18,6 +18,7 @@ from types import ModuleType
 from typing import (
     Any, Callable, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Type, TypeVar, Union
 )
+from typing import get_type_hints
 
 from docutils.statemachine import StringList
 
@@ -33,7 +34,9 @@ from sphinx.pycode import ModuleAnalyzer, PycodeError
 from sphinx.util import inspect
 from sphinx.util import logging
 from sphinx.util.docstrings import extract_metadata, prepare_docstring
-from sphinx.util.inspect import getdoc, object_description, safe_getattr, stringify_signature
+from sphinx.util.inspect import (
+    evaluate_signature, getdoc, object_description, safe_getattr, stringify_signature
+)
 from sphinx.util.typing import stringify as stringify_typehint
 
 if False:
@@ -59,13 +62,29 @@ py_ext_sig_re = re.compile(
            (?:\s* -> \s* (.*))?  #           return annotation
           )? $                   # and nothing more
           ''', re.VERBOSE)
+special_member_re = re.compile(r'^__\S+__$')
 
 
 def identity(x: Any) -> Any:
     return x
 
 
-ALL = object()
+class _All:
+    """A special value for :*-members: that matches to any member."""
+
+    def __contains__(self, item: Any) -> bool:
+        return True
+
+
+class _Empty:
+    """A special value for :exclude-members: that never matches to any member."""
+
+    def __contains__(self, item: Any) -> bool:
+        return False
+
+
+ALL = _All()
+EMPTY = _Empty()
 UNINITIALIZED_ATTR = object()
 INSTANCEATTR = object()
 SLOTSATTR = object()
@@ -80,11 +99,20 @@ def members_option(arg: Any) -> Union[object, List[str]]:
 
 def members_set_option(arg: Any) -> Union[object, Set[str]]:
     """Used to convert the :members: option to auto directives."""
+    warnings.warn("members_set_option() is deprecated.",
+                  RemovedInSphinx50Warning, stacklevel=2)
     if arg is None:
         return ALL
     return {x.strip() for x in arg.split(',') if x.strip()}
 
 
+def exclude_members_option(arg: Any) -> Union[object, Set[str]]:
+    """Used to convert the :exclude-members: option."""
+    if arg is None:
+        return EMPTY
+    return {x.strip() for x in arg.split(',') if x.strip()}
+
+
 def inherited_members_option(arg: Any) -> Union[object, Set[str]]:
     """Used to convert the :members: option to auto directives."""
     if arg is None:
@@ -123,6 +151,8 @@ def bool_option(arg: Any) -> bool:
 
 def merge_special_members_option(options: Dict) -> None:
     """Merge :special-members: option to :members: option."""
+    warnings.warn("merge_special_members_option() is deprecated.",
+                  RemovedInSphinx50Warning, stacklevel=2)
     if 'special-members' in options and options['special-members'] is not ALL:
         if options.get('members') is ALL:
             pass
@@ -134,6 +164,20 @@ def merge_special_members_option(options: Dict) -> None:
             options['members'] = options['special-members']
 
 
+def merge_members_option(options: Dict) -> None:
+    """Merge :*-members: option to the :members: option."""
+    if options.get('members') is ALL:
+        # merging is not needed when members: ALL
+        return
+
+    members = options.setdefault('members', [])
+    for key in {'private-members', 'special-members'}:
+        if key in options and options[key] is not ALL:
+            for member in options[key]:
+                if member not in members:
+                    members.append(member)
+
+
 # Some useful event listener factories for autodoc-process-docstring.
 
 def cut_lines(pre: int, post: int = 0, what: str = None) -> Callable:
@@ -334,7 +378,7 @@ class Documenter:
                         ('.' + '.'.join(self.objpath) if self.objpath else '')
         return True
 
-    def import_object(self) -> bool:
+    def import_object(self, raiseerror: bool = False) -> bool:
         """Import the object given by *self.modname* and *self.objpath* and set
         it as *self.object*.
 
@@ -348,9 +392,12 @@ class Documenter:
                 self.module, self.parent, self.object_name, self.object = ret
                 return True
             except ImportError as exc:
-                logger.warning(exc.args[0], type='autodoc', subtype='import_object')
-                self.env.note_reread()
-                return False
+                if raiseerror:
+                    raise
+                else:
+                    logger.warning(exc.args[0], type='autodoc', subtype='import_object')
+                    self.env.note_reread()
+                    return False
 
     def get_real_modname(self) -> str:
         """Get the real module name of an object to document.
@@ -574,6 +621,8 @@ class Documenter:
                         return True
                     elif name in cls.__dict__:
                         return False
+                    elif name in self.get_attr(cls, '__annotations__', {}):
+                        return False
 
             return False
 
@@ -624,31 +673,40 @@ class Documenter:
             if safe_getattr(member, '__sphinx_mock__', False):
                 # mocked module or object
                 pass
-            elif want_all and membername.startswith('__') and \
-                    membername.endswith('__') and len(membername) > 4:
+            elif self.options.exclude_members and membername in self.options.exclude_members:
+                # remove members given by exclude-members
+                keep = False
+            elif want_all and special_member_re.match(membername):
                 # special __methods__
-                if self.options.special_members is ALL:
+                if self.options.special_members and membername in self.options.special_members:
                     if membername == '__doc__':
                         keep = False
                     elif is_filtered_inherited_member(membername):
                         keep = False
                     else:
                         keep = has_doc or self.options.undoc_members
-                elif self.options.special_members:
-                    if membername in self.options.special_members:
-                        keep = has_doc or self.options.undoc_members
+                else:
+                    keep = False
             elif (namespace, membername) in attr_docs:
                 if want_all and isprivate:
-                    # ignore members whose name starts with _ by default
-                    keep = self.options.private_members
+                    if self.options.private_members is None:
+                        keep = False
+                    else:
+                        keep = membername in self.options.private_members
                 else:
                     # keep documented attributes
                     keep = True
                 isattr = True
             elif want_all and isprivate:
-                # ignore members whose name starts with _ by default
-                keep = self.options.private_members and \
-                    (has_doc or self.options.undoc_members)
+                if has_doc or self.options.undoc_members:
+                    if self.options.private_members is None:
+                        keep = False
+                    elif is_filtered_inherited_member(membername):
+                        keep = False
+                    else:
+                        keep = membername in self.options.private_members
+                else:
+                    keep = False
             else:
                 if self.options.members is ALL and is_filtered_inherited_member(membername):
                     keep = False
@@ -693,16 +751,6 @@ class Documenter:
         # find out which members are documentable
         members_check_module, members = self.get_object_members(want_all)
 
-        # remove members given by exclude-members
-        if self.options.exclude_members:
-            members = [
-                (membername, member) for (membername, member) in members
-                if (
-                    self.options.exclude_members is ALL or
-                    membername not in self.options.exclude_members
-                )
-            ]
-
         # document non-skipped members
         memberdocumenters = []  # type: List[Tuple[Documenter, bool]]
         for (mname, member, isattr) in self.filter_members(members, want_all):
@@ -859,14 +907,14 @@ class ModuleDocumenter(Documenter):
         'noindex': bool_option, 'inherited-members': inherited_members_option,
         'show-inheritance': bool_option, 'synopsis': identity,
         'platform': identity, 'deprecated': bool_option,
-        'member-order': member_order_option, 'exclude-members': members_set_option,
-        'private-members': bool_option, 'special-members': members_option,
+        'member-order': member_order_option, 'exclude-members': exclude_members_option,
+        'private-members': members_option, 'special-members': members_option,
         'imported-members': bool_option, 'ignore-module-all': bool_option
     }  # type: Dict[str, Callable]
 
     def __init__(self, *args: Any) -> None:
         super().__init__(*args)
-        merge_special_members_option(self.options)
+        merge_members_option(self.options)
         self.__all__ = None
 
     @classmethod
@@ -890,7 +938,7 @@ class ModuleDocumenter(Documenter):
                            type='autodoc')
         return ret
 
-    def import_object(self) -> Any:
+    def import_object(self, raiseerror: bool = False) -> bool:
         def is_valid_module_all(__all__: Any) -> bool:
             """Check the given *__all__* is valid for a module."""
             if (isinstance(__all__, (list, tuple)) and
@@ -899,7 +947,7 @@ class ModuleDocumenter(Documenter):
             else:
                 return False
 
-        ret = super().import_object()
+        ret = super().import_object(raiseerror)
 
         if not self.options.ignore_module_all:
             __all__ = getattr(self.object, '__all__', None)
@@ -1204,7 +1252,9 @@ class FunctionDocumenter(DocstringSignatureMixin, ModuleLevelDocumenter):  # typ
                     documenter.objpath = [None]
                     sigs.append(documenter.format_signature())
         if overloaded:
+            __globals__ = safe_getattr(self.object, '__globals__', {})
             for overload in self.analyzer.overloads.get('.'.join(self.objpath)):
+                overload = evaluate_signature(overload, __globals__)
                 sig = stringify_signature(overload, **kwargs)
                 sigs.append(sig)
 
@@ -1277,8 +1327,8 @@ class ClassDocumenter(DocstringSignatureMixin, ModuleLevelDocumenter):  # type:
         'members': members_option, 'undoc-members': bool_option,
         'noindex': bool_option, 'inherited-members': inherited_members_option,
         'show-inheritance': bool_option, 'member-order': member_order_option,
-        'exclude-members': members_set_option,
-        'private-members': bool_option, 'special-members': members_option,
+        'exclude-members': exclude_members_option,
+        'private-members': members_option, 'special-members': members_option,
     }  # type: Dict[str, Callable]
 
     _signature_class = None  # type: Any
@@ -1286,15 +1336,15 @@ class ClassDocumenter(DocstringSignatureMixin, ModuleLevelDocumenter):  # type:
 
     def __init__(self, *args: Any) -> None:
         super().__init__(*args)
-        merge_special_members_option(self.options)
+        merge_members_option(self.options)
 
     @classmethod
     def can_document_member(cls, member: Any, membername: str, isattr: bool, parent: Any
                             ) -> bool:
         return isinstance(member, type)
 
-    def import_object(self) -> Any:
-        ret = super().import_object()
+    def import_object(self, raiseerror: bool = False) -> bool:
+        ret = super().import_object(raiseerror)
         # if the class is documented under another name, document it
         # as data/attribute
         if ret:
@@ -1403,7 +1453,11 @@ class ClassDocumenter(DocstringSignatureMixin, ModuleLevelDocumenter):  # type:
         sigs = []
         if overloaded:
             # Use signatures for overloaded methods instead of the implementation method.
+            method = safe_getattr(self._signature_class, self._signature_method_name, None)
+            __globals__ = safe_getattr(method, '__globals__', {})
             for overload in self.analyzer.overloads.get(qualname):
+                overload = evaluate_signature(overload, __globals__)
+
                 parameters = list(overload.parameters.values())
                 overload = overload.replace(parameters=parameters[1:],
                                             return_annotation=Parameter.empty)
@@ -1552,8 +1606,12 @@ class DataDocumenter(ModuleLevelDocumenter):
         sourcename = self.get_sourcename()
         if not self.options.annotation:
             # obtain annotation for this data
-            annotations = getattr(self.parent, '__annotations__', {})
-            if annotations and self.objpath[-1] in annotations:
+            try:
+                annotations = get_type_hints(self.parent)
+            except TypeError:
+                annotations = {}
+
+            if self.objpath[-1] in annotations:
                 objrepr = stringify_typehint(annotations.get(self.objpath[-1]))
                 self.add_line('   :type: ' + objrepr, sourcename)
             else:
@@ -1604,7 +1662,7 @@ class DataDeclarationDocumenter(DataDocumenter):
                 isattr and
                 member is INSTANCEATTR)
 
-    def import_object(self) -> bool:
+    def import_object(self, raiseerror: bool = False) -> bool:
         """Never import anything."""
         # disguise as a data
         self.objtype = 'data'
@@ -1703,8 +1761,8 @@ class MethodDocumenter(DocstringSignatureMixin, ClassLevelDocumenter):  # type:
         return inspect.isroutine(member) and \
             not isinstance(parent, ModuleDocumenter)
 
-    def import_object(self) -> Any:
-        ret = super().import_object()
+    def import_object(self, raiseerror: bool = False) -> bool:
+        ret = super().import_object(raiseerror)
         if not ret:
             return ret
 
@@ -1796,7 +1854,9 @@ class MethodDocumenter(DocstringSignatureMixin, ClassLevelDocumenter):  # type:
                     documenter.objpath = [None]
                     sigs.append(documenter.format_signature())
         if overloaded:
+            __globals__ = safe_getattr(self.object, '__globals__', {})
             for overload in self.analyzer.overloads.get('.'.join(self.objpath)):
+                overload = evaluate_signature(overload, __globals__)
                 if not inspect.isstaticmethod(self.object, cls=self.parent,
                                               name=self.object_name):
                     parameters = list(overload.parameters.values())
@@ -1869,15 +1929,42 @@ class AttributeDocumenter(DocstringStripSignatureMixin, ClassLevelDocumenter):
     def document_members(self, all_members: bool = False) -> None:
         pass
 
-    def import_object(self) -> Any:
-        ret = super().import_object()
-        if inspect.isenumattribute(self.object):
-            self.object = self.object.value
-        if inspect.isattributedescriptor(self.object):
-            self._datadescriptor = True
-        else:
-            # if it's not a data descriptor
-            self._datadescriptor = False
+    def isinstanceattribute(self) -> bool:
+        """Check the subject is an instance attribute."""
+        try:
+            analyzer = ModuleAnalyzer.for_module(self.modname)
+            attr_docs = analyzer.find_attr_docs()
+            if self.objpath:
+                key = ('.'.join(self.objpath[:-1]), self.objpath[-1])
+                if key in attr_docs:
+                    return True
+
+            return False
+        except PycodeError:
+            return False
+
+    def import_object(self, raiseerror: bool = False) -> bool:
+        try:
+            ret = super().import_object(raiseerror=True)
+            if inspect.isenumattribute(self.object):
+                self.object = self.object.value
+            if inspect.isattributedescriptor(self.object):
+                self._datadescriptor = True
+            else:
+                # if it's not a data descriptor
+                self._datadescriptor = False
+        except ImportError as exc:
+            if self.isinstanceattribute():
+                self.object = INSTANCEATTR
+                self._datadescriptor = False
+                ret = True
+            elif raiseerror:
+                raise
+            else:
+                logger.warning(exc.args[0], type='autodoc', subtype='import_object')
+                self.env.note_reread()
+                ret = False
+
         return ret
 
     def get_real_modname(self) -> str:
@@ -1889,8 +1976,12 @@ class AttributeDocumenter(DocstringStripSignatureMixin, ClassLevelDocumenter):
         sourcename = self.get_sourcename()
         if not self.options.annotation:
             # obtain type annotation for this attribute
-            annotations = getattr(self.parent, '__annotations__', {})
-            if annotations and self.objpath[-1] in annotations:
+            try:
+                annotations = get_type_hints(self.parent)
+            except TypeError:
+                annotations = {}
+
+            if self.objpath[-1] in annotations:
                 objrepr = stringify_typehint(annotations.get(self.objpath[-1]))
                 self.add_line('   :type: ' + objrepr, sourcename)
             else:
@@ -1984,11 +2075,22 @@ class InstanceAttributeDocumenter(AttributeDocumenter):
                 isattr and
                 member is INSTANCEATTR)
 
-    def import_object(self) -> bool:
+    def import_parent(self) -> Any:
+        try:
+            parent = importlib.import_module(self.modname)
+            for name in self.objpath[:-1]:
+                parent = self.get_attr(parent, name)
+
+            return parent
+        except (ImportError, AttributeError):
+            return None
+
+    def import_object(self, raiseerror: bool = False) -> bool:
         """Never import anything."""
         # disguise as an attribute
         self.objtype = 'attribute'
         self.object = INSTANCEATTR
+        self.parent = self.import_parent()
         self._datadescriptor = False
         return True
 
@@ -2015,7 +2117,7 @@ class SlotsAttributeDocumenter(AttributeDocumenter):
         """This documents only SLOTSATTR members."""
         return member is SLOTSATTR
 
-    def import_object(self) -> Any:
+    def import_object(self, raiseerror: bool = False) -> bool:
         """Never import anything."""
         # disguise as an attribute
         self.objtype = 'attribute'
@@ -2029,9 +2131,12 @@ class SlotsAttributeDocumenter(AttributeDocumenter):
                 self.module, _, _, self.parent = ret
                 return True
             except ImportError as exc:
-                logger.warning(exc.args[0], type='autodoc', subtype='import_object')
-                self.env.note_reread()
-                return False
+                if raiseerror:
+                    raise
+                else:
+                    logger.warning(exc.args[0], type='autodoc', subtype='import_object')
+                    self.env.note_reread()
+                    return False
 
     def get_doc(self, encoding: str = None, ignore: int = None) -> List[List[str]]:
         """Decode and return lines of the docstring(s) for the object."""
